/*
 * This file is part of the AdvanceMAME project.
 *
 * Copyright (C) 1999-2002 Andrea Mazzoleni
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * In addition, as a special exception, Andrea Mazzoleni
 * gives permission to link the code of this program with
 * the MAME library (or with modified versions of MAME that use the
 * same license as MAME), and distribute linked combinations including
 * the two.  You must obey the GNU General Public License in all
 * respects for all of the code used other than MAME.  If you modify
 * this file, you may extend this exception to your version of the
 * file, but you are not obligated to do so.  If you do not wish to
 * do so, delete this exception statement from your version.
 */

#ifndef __CRTC_H
#define __CRTC_H

#include "monitor.h"

#ifdef __cplusplus
extern "C" {
#endif

typedef struct video_crtc_struct {
	unsigned hde,hrs,hre,ht; /**< Horizontal CRTC values, normalized 1 unit==1 pixel */
	unsigned vde,vrs,vre,vt; /**< Vertical CRTC values, normalized 1 unit==1 pixel (also for doublescan and interlaced modes) */

	/**
	 * Pixel clock. Ignore doublescan and interlaced. Tthis imply that
	 * this value is doubled if doublescan is active, or halved if interlaced
	 */
	unsigned pixelclock;

	int doublescan; /**< Double scan mode, 0==no,1==yes */
	int vsync; /**< Horizontal sync polarity, 1==negative,0==positive */
	int hsync; /**< Horizontal sync polarity, 1==negative,0==positive */
	int interlaced; /**< Interlaced mode, 0==no,1==yes */

	char name[VIDEO_NAME_MAX]; /**< Name */
	struct video_crtc_struct* container_next; /**< Used by the container */
	unsigned user_flags; /**< Flags for the user */
} video_crtc;

/* Allowed steps in CRTC values */
#define CRTC_HSTEP 8
#define CRTC_VSTEP 1

unsigned crtc_step(double v, unsigned st);

video_error crtc_adjust_clock(video_crtc* crtc, const video_monitor* monitor);

#define CRTC_ADJUST_EXACT 0x0001 /* only the exact mod */
#define CRTC_ADJUST_VCLOCK 0x0002 /* also mode with different vclock */
#define CRTC_ADJUST_VTOTAL 0x0004 /* also mode with different vtotal */
video_error crtc_find(unsigned* req_vtotal, double* req_vclock, double* req_factor, const video_monitor* monitor, unsigned cap, unsigned adjust);

double crtc_hclock_get(const video_crtc* crtc);
double crtc_vclock_get(const video_crtc* crtc);

static __inline__  double crtc_pclock_get(const video_crtc* crtc) {
	return crtc->pixelclock;
}
void crtc_vclock_set(video_crtc* crtc, double vclock);
void crtc_hclock_set(video_crtc* crtc, double hclock);
void crtc_pclock_set(video_crtc* crtc, double pclock);
void crtc_hsize_set(video_crtc* crtc, unsigned hsize);
void crtc_vsize_set(video_crtc* crtc, unsigned vsize);

static __inline__  const char* crtc_name_get(const video_crtc* crtc) {
	return crtc->name;
}

static __inline__  unsigned crtc_hsize_get(const video_crtc* crtc) {
	return crtc->hde;
}

static __inline__  unsigned crtc_vsize_get(const video_crtc* crtc) {
	return crtc->vde;
}

video_bool crtc_clock_check(const video_monitor* monitor, const video_crtc* crtc);

int video_crtc_compare(const video_crtc* a,const video_crtc* b);

video_error video_crtc_parse(video_crtc* crtc, const char* begin, const char* end);
void video_crtc_print(char* buffer, const video_crtc* crtc);

#ifdef __cplusplus
}
#endif

#endif
